/*
 * Copyright (c) 2004, Accent .
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY Accent ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 *
* This file is part of the lwIP TCP/IP stack.^M
 * 
 * Author: Alessandro Timossi <alessandro.timossi@accent.it>
 *
 */


#include <stdio.h>
#include <string.h>
#include "lwip/debug.h"
#include "lwip/stats.h"
#include "lwip/mem.h"
#include "lwip/udp.h"
#include "lwip/netif.h"
#include "lwip/dhcp.h"
#include "lwip/inet.h"
#include "bootp.h"
#include "netif/etharp.h"
#include "lwip/udp.h"
#include "lwip/ip_addr.h"
#include "lwip/sys.h"
#include "lwip/opt.h"




/* global transaction identifier, must be unique for each BOOTP request. */
static u32_t xid = XID_VALUE;

//extern struct netif *netif;


/*=================================================================
 * PROTOTYPE DECLARATIONS
=================================================================*/
static err_t bootp_create_request(struct netif *netif);
extern err_t bootp_start(struct netif *netif);
static err_t bootp_send_request(struct netif *netif);
static void bootp_handle_reply(struct netif *netif);
static void bootp_recv(void *arg, struct udp_pcb *pcb, struct pbuf *p, struct ip_addr *addr, u16_t port);
static void bootp_bind(struct netif *netif);
static err_t bootp_unfold_reply(struct bootp *bootp);
err_t tftp_rrq_send(struct netif *netif);
static void tftp_recv(void *arg, struct udp_pcb *pcb, struct pbuf *p, struct ip_addr *addr, u16_t port);
int StringToExa(__packed u8_t exaString[],int exaStringSize, char *string);



/*=================================================================
 * FUNCTION: bootp_create_request
 =================================================================*/
/**
 * DESCRIPTION: creates request packet for BOOTP request
 *
 * @param netif a pre-allocated network interface structure
 *
 * @return nothing
 */
/*================================================================*/
static err_t bootp_create_request(struct netif *netif)
{
  struct bootp *bootp = netif->bootp;
  u16_t i;
  
		bootp->p_out = pbuf_alloc(PBUF_TRANSPORT, sizeof(struct bootp_msg), PBUF_RAM);
  
		if (bootp->p_out == NULL) {
    return ERR_MEM;
  }
  /* give unique transaction identifier to this request */
  bootp->xid = xid++;

  bootp->msg_out = (struct bootp_msg *)bootp->p_out->payload;

  bootp->msg_out->op = DHCP_BOOTREQUEST;
  bootp->msg_out->htype = DHCP_HTYPE_ETH;
  bootp->msg_out->hlen = DHCP_HLEN_ETH;
  bootp->msg_out->hops = 0;
  bootp->msg_out->xid = htonl(bootp->xid);
  bootp->msg_out->secs = 0;
  bootp->msg_out->flags = 0x0080;
  bootp->msg_out->ciaddr = netif->ip_addr.addr;
  bootp->msg_out->yiaddr = 0;
  bootp->msg_out->siaddr = 0;
  bootp->msg_out->giaddr = 0;
  
		for (i = 0; i < DHCP_CHADDR_LEN; i++) {
    /* copy netif hardware address, pad with zeroes */
    bootp->msg_out->chaddr[i] = (i < netif->hwaddr_len) ? netif->hwaddr[i] : 0/* pad byte*/;
  }
  
		/* fill server name */
		StringToExa(bootp->msg_out->sname,sizeof( ((struct bootp*)bootp)->msg_out->sname),BOOTP_SNAME);
  
		/* fill boot file name */
		StringToExa(bootp->msg_out->file,sizeof( ((struct bootp*)bootp)->msg_out->file),BOOTP_FILE);
  
		/* fill magic cookie */
		bootp->msg_out->cookie = htonl(0x63825363UL);
  
		/* fill options field with an 0 array (for debugging purposes) */
  for (i = 0; i < 60; i++) bootp->msg_out->vend[i] = 0;
  return ERR_OK;
}





/*=================================================================
 * FUNCTION: bootp_start
 =================================================================*/
/**
 * DESCRIPTION: starts BOOTP process
 *
 * @param netif a pre-allocated network interface structure
 *
 * @return nothing
 */
/*================================================================*/
err_t bootp_start(struct netif *netif)
{
  struct bootp *bootp = netif->bootp;
  err_t result = ERR_OK;


  if (bootp == NULL) {
    bootp = mem_malloc(sizeof(struct bootp));
    if (bootp == NULL) {
      netif->flags &= ~NETIF_FLAG_DHCP;
      return ERR_MEM;
    }
    /* clear data structure */
    memset(bootp, 0, sizeof(struct bootp));
				
				/* Assigning offset for TFTP receiving data buffer */
		  bootp->tftp_data_buffer = (u8_t *)TFTP_DATA_POINTER;
		  /* Assigning initial value for TFTP receiving data buffer index */
		  bootp->tftp_data_buffer_index = 0;
		  /* Resetting the flag for tftp end of transfer */
		  bootp->tftp_end = FALSE;
				
    bootp->pcb = udp_new();
    if (bootp->pcb == NULL) {
      LWIP_DEBUGF(DHCP_DEBUG  | DBG_TRACE, ("dhcp_start(): could not obtain pcb\n"));
      mem_free((void *)bootp);
      bootp = NULL;
      netif->flags &= ~NETIF_FLAG_DHCP;
      return ERR_MEM;
    }
    /* store this dhcp client in the netif */
    netif->bootp = bootp;
  } else {
    LWIP_DEBUGF(DHCP_DEBUG | DBG_TRACE | DBG_STATE | 3, ("dhcp_start(): restarting DHCP configuration\n"));
  }
  /* (re)start the DHCP negotiation */
  result = bootp_send_request(netif);

  return result;
}



/*=================================================================
 * FUNCTION: bootp_send_request
 =================================================================*/
/**
 * DESCRIPTION: sends BOOTP request packet
 *
 * @param netif a pre-allocated network interface structure
 *
 * @return nothing
 */
/*================================================================*/
static err_t bootp_send_request(struct netif *netif)
{
  struct bootp *bootp = netif->bootp;
  err_t result = ERR_OK;

  ip_addr_set(&bootp->offered_ip_addr, IP_ADDR_ANY);
  
		/* create and initialize the DHCP message header */
  result = bootp_create_request(netif);
  if (result == ERR_OK)
  {

    /* set receive callback function with netif as user data */
    udp_recv(bootp->pcb, bootp_recv, netif);
    
				udp_bind(bootp->pcb, IP_ADDR_ANY, DHCP_CLIENT_PORT);
    
				udp_connect(bootp->pcb, IP_ADDR_ANY, DHCP_SERVER_PORT);
    
				udp_sendto(bootp->pcb, bootp->p_out, IP_ADDR_BROADCAST, DHCP_SERVER_PORT);
    
  } else {
    LWIP_DEBUGF(DHCP_DEBUG | DBG_TRACE | 2, ("bootp_send_request: error on BOOTP request\n"));
  }
  
		return result;
}




/*=================================================================
 * FUNCTION: bootp_handle_reply
 =================================================================*/
/**
 * DESCRIPTION: handles the BOOTP reply from bootp server
 *
 * @param netif a pre-allocated network interface structure
 *
 * @return nothing
 */
/*================================================================*/
static void bootp_handle_reply(struct netif *netif)
{
  struct bootp *bootp = netif->bootp;
  
  /* remember offered address */
  ip_addr_set(&bootp->offered_ip_addr, (struct ip_addr *)&bootp->msg_in->yiaddr);
}



/*=================================================================
 * FUNCTION: bootp_unfold_reply
 =================================================================*/
/**
 * DESCRIPTION: unfolds the BOOTP reply from bootp server
 *
 * @param netif a pre-allocated network interface structure
 *
 * @return nothing
 */
/*================================================================*/
static err_t bootp_unfold_reply(struct bootp *bootp)
{
  struct pbuf *p = bootp->p;
  u8_t *ptr;
  u16_t i;
  u16_t j = 0;
  
  /* options present? */
  if (bootp->p->tot_len > (sizeof(struct bootp_msg) - 60))
  {
    bootp->options_in_len = bootp->p->tot_len - (sizeof(struct bootp_msg) - 60);
    bootp->options_in = mem_malloc(bootp->options_in_len);
    if (bootp->options_in == NULL)
    {
      return ERR_MEM;
    }
  }
  bootp->msg_in = mem_malloc(sizeof(struct bootp_msg) - 60);
  if (bootp->msg_in == NULL)
  {
    mem_free((void *)bootp->options_in);
    bootp->options_in = NULL;
    return ERR_MEM;
  }

  ptr = (u8_t *)bootp->msg_in;
  /* proceed through struct dhcp_msg */
  for (i = 0; i < sizeof(struct bootp_msg) - 60; i++)
  {
    *ptr++ = ((u8_t *)p->payload)[j++];
    /* reached end of pbuf? */
    if (j == p->len)
    {
      /* proceed to next pbuf in chain */
      p = p->next;
      j = 0;
    }
  }
  if (bootp->options_in != NULL) {
    ptr = (u8_t *)bootp->options_in;
    /* proceed through options */
    for (i = 0; i < bootp->options_in_len; i++) {
      *ptr++ = ((u8_t *)p->payload)[j++];
      /* reached end of pbuf? */
      if (j == p->len) {
        /* proceed to next pbuf in chain */
        p = p->next;
        j = 0;
      }
    }
  }
  return ERR_OK;
}



/*=================================================================
 * FUNCTION: bootp_recv
 =================================================================*/
/**
 * DESCRIPTION: registered function for managing bootp rx process
 *
 * @param arg a pre-allocated network interface structure
 * @param pcb protocol control block structure
 * @param p as pbuf pointer
 * @param local ip address
 * @param local port
 *
 * @return nothing
 */
/*================================================================*/
static void bootp_recv(void *arg, struct udp_pcb *pcb, struct pbuf *p, struct ip_addr *addr, u16_t port)
{
  struct netif *netif = (struct netif *)arg;
  struct bootp *bootp = netif->bootp;
  struct bootp_msg *reply_msg = (struct bootp_msg *)p->payload;
  u8_t i;
  
  /* prevent warnings about unused arguments */
  (void)pcb; (void)addr; (void)port;
  
		bootp->p = p;
  
		/* TODO: check packet length before reading them */
  if (reply_msg->op != DHCP_BOOTREPLY) {
    pbuf_free(p);
    bootp->p = NULL;
    return;
  }
  /* iterate through hardware address and match against DHCP message */
  for (i = 0; i < netif->hwaddr_len; i++) {
    if (netif->hwaddr[i] != reply_msg->chaddr[i]) {
      pbuf_free(p);
      bootp->p = NULL;
      return;
    }
  }
  /* match transaction ID against what we expected */
  if (ntohl(reply_msg->xid) != bootp->xid) {
    pbuf_free(p);
    bootp->p = NULL;
    return;
  }

  /* option fields could be unfold? */
  if (bootp_unfold_reply(bootp) != ERR_OK) {
    pbuf_free(p);
    bootp->p = NULL;
    return;
  }


  bootp_handle_reply(netif);
		
		bootp_bind(netif);
  
		pbuf_free(p);
  bootp->p = NULL;
		
		tftp_rrq_send(netif);
				
}



/*=================================================================
 * FUNCTION: bootp_bind
 =================================================================*/
/**
 * DESCRIPTION: binds the network interface to the received address
 *
 * @param netif a pre-allocated network interface structure
 *
 * @return nothing
 */
/*================================================================*/
static void bootp_bind(struct netif *netif)
{
  struct bootp *bootp = netif->bootp;
  struct ip_addr sn_mask, gw_addr;


  /* copy offered network mask */
  ip_addr_set(&sn_mask, &bootp->offered_sn_mask);

  /* subnet mask not given? */
  /* TODO: this is not a valid check. what if the network mask is 0? */
  if (sn_mask.addr == 0) {
    /* choose a safe subnet mask given the network class */
    u8_t first_octet = ip4_addr1(&sn_mask);
    if (first_octet <= 127) sn_mask.addr = htonl(0xff000000);
    else if (first_octet >= 192) sn_mask.addr = htonl(0xffffff00);
    else sn_mask.addr = htonl(0xffff0000);
  }

  ip_addr_set(&gw_addr, &bootp->offered_gw_addr);
  /* gateway address not given? */
  if (gw_addr.addr == 0) {
    /* copy network address */
    gw_addr.addr = (bootp->offered_ip_addr.addr & sn_mask.addr);
    /* use first host address on network as gateway */
    gw_addr.addr |= htonl(0x00000001);
  }

  netif_set_ipaddr(netif, &bootp->offered_ip_addr);
  netif_set_netmask(netif, &sn_mask);
  netif_set_gw(netif, &gw_addr);

}



/*=================================================================
 * FUNCTION: tftp_rrq_send
 =================================================================*/
/**
 * DESCRIPTION: sends TFTP RRQ request packet to start TFTP read
 *
 * @param netif a pre-allocated network interface structure
 *
 * @return nothing
 */
/*================================================================*/
err_t tftp_rrq_send(struct netif *netif)
{
  struct pbuf *p;
		struct bootp *bootp = netif->bootp; 
  struct tftp_RRQ_packet *packet;
		struct ip_addr *ipaddr;
  u32_t i;
 
	 /* Allocating pbuf for TFTP RRQ packet */
  p = pbuf_alloc(PBUF_TRANSPORT, sizeof(struct tftp_RRQ_packet), PBUF_RAM); 
   if(p == NULL) {
      return ERR_MEM;
    }
		
  packet = p->payload;

  
		/* Assigning request OPCODE */
  packet->opcode = REQUEST;
  
		/* Assigning FILENAME */
		//for (i = 0; i < DHCP_FILE_LEN; i++) packet->filename[i] = bootp->msg_in->file[i];
		StringToExa(packet->filename,sizeof( ((struct tftp_RRQ_packet*)packet)->filename),BOOTP_FILE);
		
		/* Assigning MODE */
  StringToExa(packet->mode,sizeof( ((struct tftp_RRQ_packet*)packet)->mode),BOOTP_MODE);
		
		/* Assigning dummy bytes */
		packet->dummy1 = 0;
		packet->dummy2 = 0;
		packet->dummy3 = 0;
		packet->dummy4 = 0;
		
		/* Assigning option bytes */
		StringToExa(packet->opt1,sizeof( ((struct tftp_RRQ_packet*)packet)->opt1),OPT1);
		
		/* Assigning option value bytes */
		StringToExa(packet->value1,sizeof( ((struct tftp_RRQ_packet*)packet)->value1),VALUE1); 
		
		
		/* Starting TFTP session */
		bootp->tftp_pcb = udp_new();
		
		udp_recv(bootp->tftp_pcb, tftp_recv, netif);
    
		udp_bind(bootp->tftp_pcb, &netif->ip_addr, xid);
		
		ipaddr->addr = bootp->msg_in->siaddr;
    
		udp_connect(bootp->tftp_pcb, ipaddr, 69);
        		
		udp_sendto(bootp->tftp_pcb, p, ipaddr, 69);

  return ERR_OK;
}



/*=================================================================
 * FUNCTION: tftp_recv
 =================================================================*/
/**
 * DESCRIPTION: registered function for managing tftp rx process
 *
 * @param arg a pre-allocated network interface structure
 * @param pcb protocol control block structure
 * @param p as pbuf pointer
 * @param local ip address
 * @param local port
 *
 * @return nothing
 */
/*================================================================*/
static void tftp_recv(void *arg, struct udp_pcb *pcb, struct pbuf *p, struct ip_addr *addr, u16_t port)
{
  struct netif *netif = (struct netif *)arg;
  struct pbuf *ack_pbuf;
		struct bootp *bootp = netif->bootp; 
  struct tftp_DATA_packet *data_packet;
		struct tftp_ACK_packet *ack_packet;
		struct ip_addr *ipaddr; 
  u32_t i;
  
		/* prevent warnings about unused arguments */
  (void)pcb; (void)addr; (void)port;
		
		
		/* The tftp_DATA_packet points to the currently received pbuf payload */
		data_packet = p->payload;	

  /* Verifying request OPCODE */
  if(data_packet->opcode != DATA){
		  pbuf_free(p);
    p = NULL;
    return;
		 }
	

  /* Storing TFTP DATA packet */
		for(i=0; i<p->len-4; i++){
		   bootp->tftp_data_buffer[bootp->tftp_data_buffer_index] = data_packet->data[i];
					bootp->tftp_data_buffer_index++;
		  }
				
		/* Report if TFTP DATA transfer has been finished */
		if(p->len-4 < MAX_TFTP_LEN - 1)
		    bootp->tftp_end = TRUE;

	 /* Preparing pbuf for TFTP ACK packet */
  ack_pbuf = pbuf_alloc(PBUF_TRANSPORT, sizeof(struct tftp_ACK_packet), PBUF_RAM); 
   if(ack_pbuf == NULL) {
      return;
    } 
	 
		/* Filling ACK packet pbuf */
		ack_packet = ack_pbuf->payload;
  
		/* Assigning ACK packet bytes */
		ack_packet->opcode = ACK;
		ack_packet->block = data_packet->block;
		
  /* Sending ACK packet bytes */      
  ipaddr->addr = bootp->msg_in->siaddr;
		
		bootp->tftp_pcb->local_ip.addr = netif->ip_addr.addr;
		
		udp_sendto(bootp->tftp_pcb, ack_pbuf, ipaddr, 69);

}




/*=================================================================
 * FUNCTION: StringToExa
 =================================================================*/
/**
 * DESCRIPTION: converts a defined string to a byte array
 *
 * @param exaString: array to be filled
 * @param exaStringSize: size of array to be filled
 * @param string: string to be converted	
 *
 * @return nothing
 */
/*================================================================*/
int StringToExa(__packed u8_t exaString[],int exaStringSize, char *string)
{
   int i,len;
   
   len = strlen(string);
 
   if ( exaStringSize < len)
      return (-1);
   for (i=0; i < len; i++)
   {
      exaString[i] = *string;
	  string++;
   }
   
   for (i=len; i<exaStringSize; i++)
   {
      exaString[i] = '\0';
   }  
   
   return len;

}
